#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <stdarg.h>
#include <fcntl.h>
#include <fnmatch.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <sys/vfs.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include "dftpd.h"

/* libdftp dependencies */

char *dftp_printbytes(unsigned long long);

/* dftpd.c dependencies */

extern struct dftp_config cfg;
extern int sockfd, dupefile;
extern struct dftp_vars var;
extern struct dftp_utmpent cur_utmp;
extern struct sockaddr_in cli_addr, data_addr;

void ftplog(const char *, const char *, ...);
void closefiles();

/* cmd.c dependencies */

extern unsigned long long sessiondl, sessionul;

/* srv-common.c dependencies */

char *getident();
void writes(const char *,...);

unsigned long long freespace()
{
   struct statfs puskuri;

   int fd = open(".",0); fstatfs(fd,&puskuri);
   return (unsigned long long)puskuri.f_bsize*(unsigned long long)puskuri.f_bavail;
}

short gethname(struct sockaddr_in *target, char *hostname)
{
   int i; char *osoite = NULL; struct hostent *peer, *checkpeer;

   osoite = inet_ntoa(target->sin_addr);
   if(!cfg.resolvehosts)
   {
      strncpy(hostname,osoite,HOSTNAMELEN); return 0;
   }
   peer = gethostbyaddr((char *) &target->sin_addr,sizeof(struct in_addr),AF_INET);
   if(!peer)
   {
      if(cfg.debuglevel >= 2 && h_errno == HOST_NOT_FOUND) ftplog("AUTH","DEBUG: could not resolve peer name");
      if(cfg.debuglevel >= 2 && h_errno == NO_RECOVERY) ftplog("AUTH","DEBUG: nonrecoverable name server error");
      strncpy(hostname,osoite,HOSTNAMELEN); return 2;
   }
   if((checkpeer = gethostbyname(peer->h_name)))
   {
      for(i=0;checkpeer->h_addr_list[i];i++)
         if(!strcmp(peer->h_addr_list[0],checkpeer->h_addr_list[i])) break;
      if(checkpeer->h_addr_list[i])
      {
         strncpy(hostname,peer->h_name,HOSTNAMELEN); return 0;
      }
   }
   ftplog("CONN","Warning: Peer host name(%s) seems spoofed, replacing it with ip address",peer->h_name);
   strncpy(hostname,osoite,HOSTNAMELEN); return 1;
}

void viewfile(const char *, const short);

int cookie(const char *cookie)
{
   int rvalue = 0, mod = 0;

   if(!strlen(cookie)) return 0;
   if(strlen(cookie) > 1 && cookie[0] == '-')
   {
      mod = 1; cookie++;
   }
   if(strlen(cookie) > 4 && !strncmp(cookie,"fv[",3) && strchr(cookie,']'))
   {
      int size = strchr(cookie,']')-strchr(cookie,'[')-1; char *file = malloc(size+1);
      strncpy(file,cookie+3,size); file[size] = 0;
      viewfile(file,0); rvalue = size+4;
   }
   if(rvalue) return rvalue+mod;
   rvalue = 2;
   if(strlen(cookie) > 1)
   {
      if(!strncmp(cookie,"cr",2))
      {
         if(mod) writes("%qu",var.user.credits);
         else writes("%s",dftp_printbytes(var.user.credits));
      }
      else if(cookie[0] == 'c' && isdigit(cookie[1]))
      {
         writes("%c[%c",27,cookie[1]);
         if(strlen(cookie) > 3 && isdigit(cookie[2]) && isdigit(cookie[3]))
         {
            writes(";3%c;4%c",cookie[2],cookie[3]); rvalue = 4;
         }
         writes("m"); rvalue = 2;
      }
      else if(!strncmp(cookie,"hn",2)) writes("%s",cur_utmp.controlhostname);
      else if(!strncmp(cookie,"hd",2) && var.user.homedir) writes("%s",var.user.homedir);
      else if(!strncmp(cookie,"un",2)) writes("%s",var.user.name);
      else if(!strncmp(cookie,"gn",2)) writes("%s",var.group.name);
      else if(!strncmp(cookie,"ui",2)) writes("%s",var.user.uid);
      else if(!strncmp(cookie,"gi",2)) writes("%s",var.user.gid);
      else if(!strncmp(cookie,"su",2))
      {
         if(mod) writes("%qu",var.sessionul);
         else writes("%s",dftp_printbytes(var.sessionul));
      }
      else if(!strncmp(cookie,"sd",2))
      {
         if(mod) writes("%qu",var.sessiondl);
         else writes("%s",dftp_printbytes(var.sessiondl));
      }
      else if(*cookie == 't')
      {
         char timestr[31], format[3]; time_t time1; struct tm *time2;

         time(&time1); time2 = localtime(&time1);
         sprintf(format,"%%%c",cookie[1]);
         memset(timestr,0,31); strftime(timestr,30,format,time2);
         writes("%s",timestr);
      }
      else rvalue = 0;
   }
   if(rvalue) return rvalue+mod;
   rvalue = 1;
   switch(*cookie)
   {
      case 'e': writes("%hu",var.group.id); break;
      case 'r': if(!var.user.ratio) writes("unlimited");
                else writes("%hu",var.user.ratio); break;
      case 'd': if(mod) writes("%qu",var.user.dlbytes);
                else writes("%s",dftp_printbytes(var.user.dlbytes)); break;
      case 'u': if(mod) writes("%qu",var.user.ulbytes);
                else writes("%s",dftp_printbytes(var.user.ulbytes)); break;
      case 'f': if(mod) writes("%qu",freespace());
                else writes("%s",dftp_printbytes(freespace())); break;
      case 'i': writes("%s",getident()); break;
      case 'v': writes("%s",VER); break;
      case '%': writes("%"); break;
      default:  rvalue = 0;
   }
   return rvalue+mod;
}

/* writes() with cookies enabled */

void cwrites(const char *format, ...)
{
   int len; char buf[512], *ptr = buf; va_list ap;

   memset(buf,0,512); va_start(ap,format); vsnprintf(buf,511,format,ap);
   while(strlen(ptr))
   {
      len = strcspn(ptr,"%");
      write(sockfd,ptr,len); ptr += len+1;
      if(*ptr) ptr += cookie(ptr);
   }
}

int checklog()
{
   if(var.logged < 2)
   {
      writes("530 Please login with USER and PASS.\r\n");
      return(0);
   }
   return(-1);
}

int checkaccess(char flag)
{
   int i;
   
   if(var.user.flags && strchr(var.user.flags,flag)) return 1;
   for(i=0;i<cfg.flagcount;i++) if(flag == cfg.flags[i].mark) break;
   if(i<cfg.flagcount) writes("532 %s access denied.\r\n",cfg.flags[i].name);
   else writes("532 (unknown) access denied.\r\n");
   return 0;
}

char **parsestr(char *source)
{
   char **dest = NULL, *ptr = source, *ptr2; int destc = 0;

   while((ptr2=strsep(&ptr," ")))
   {
      dest = realloc(dest,(++destc+1)*sizeof(void *));
      dest[destc-1] = malloc(strlen(ptr2)+1); strcpy(dest[destc-1],ptr2);
      dest[destc] = NULL;
   }
   return dest;
}

void _status(unsigned short luku)
{
   if(cfg.statusline == NULL) writes("%hu Transmission complete.\r\n",luku);
   else cwrites("%hu %s\r\n",luku,cfg.statusline);
}

void resetvars()
{
   memset(&var,0,sizeof(struct dftp_vars));
   var.trans_type = 'A'; var.cmode = cfg.cmode;
}

void viewfile(const char *args, const short number)
{
   int i,fd,nread; char buf[KB], cr = 0;

   errno = 0;
   fd = open(args,O_RDONLY);
   if(fd < 0) return;
   if(number > 0) writes("%hu-",number);
   while( (nread = read(fd,buf,KB)) > 0)
   {
      for(i = 0; i < nread; i++)
      {
         if(buf[i] == '\n')
         {
            cr = 1;
            if(!number)
            {
               close(fd); return;
            }
            writes("\r\n");
         }
         else
         {
            if(cr)
            {
               if(number > 0) writes("%hu-",number); else writes(" ");
            }
            writes("%c",buf[i]); cr = 0;
         }
      }
   }
   if(number && !cr) writes("\r\n"); close(fd);
}

void tildeexp(char *args)
{
   char *tildeptr, *line;

   if(!(*args) || !var.user.homedir) return;
   if(!(tildeptr = strchr(args,'~'))) return;
   line = malloc(strlen(args)+strlen(var.user.homedir));
   memcpy(line,args,tildeptr-args); line[tildeptr-args] = 0;
   strcat(line,var.user.homedir);
   strcat(line,tildeptr+1); strncpy(args,line,MAXLINE); args[MAXLINE-1] = 0;
}

void update_env()
{
   char text[20];

   setenv("USER",var.user.name,1);
   setenv("GROUP",var.group.name,1);
   sprintf(text,"%qu",var.user.dlbytes); setenv("DLBYTES",text,1);
   sprintf(text,"%qu",var.user.ulbytes); setenv("ULBYTES",text,1);
   sprintf(text,"%qu",var.sessiondl); setenv("SESSIONDL",text,1);
   sprintf(text,"%qu",var.sessionul); setenv("SESSIONUL",text,1);
   sprintf(text,"%qu",var.user.credits); setenv("CREDITS",text,1);
   if(var.user.homedir) setenv("HOME",var.user.homedir,1);
   if(var.user.flags) setenv("FLAGS",var.user.flags,1);
}
