#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <glob.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include "dftpd.h"

/* aux.c dependencies */

void cwrites(const char *, ...);
void viewfile(const char *, const short);
int checklog();
int checkaccess(int);
void update_env();
void tildeexp(char *);

/* srvcommon.c dependencies */

void writes(const char *, ...);
void readline(int, char *);

/* dftpd.c dependencies */

void ftplog(const char *, const char *, ...);
void activity(char *);

extern struct dftp_config cfg;
extern struct dftp_vars var;
extern struct sockaddr_in data_addr;
extern int sockfd, pasv_servsockfd;
extern struct sockaddr_in pasv_addr;

/* login.c dependencies */

void auth(char *);
void user(char *);
void pass(char *);

/* list.c dependencies */

void listcmd(int, char *);

/* xfer.c dependencies */

void get(char *args);
void put(char *args);

/* private variables */

char *renamefile = NULL; unsigned long offset;

void quit()
{
   shutdown(sockfd,0);
   if(!cfg.goodbyemsg) writes("220 Goodbye.\r\n");
   else cwrites("220 %s\r\n",cfg.goodbyemsg);
   ftplog("CONN","Connection closed");
   exit(0);
}

void pwd()
{
   char wd[PATH_MAX];

   if(!checklog()) return;

   getcwd(wd,PATH_MAX);
   writes("257 \"%s\" is current directory.\r\n",wd);
}

void rnfr(char *args)
{
   if(!checklog()) return;
   if(var.userclass != REAL && !checkaccess('D')) return;

   if(access(args,F_OK))
   {
      writes("550 %s: No such file or directory.\r\n",args);
      return;
   }
   if(renamefile != NULL) 
   {
      free(renamefile); renamefile = NULL;
   }
   renamefile = malloc(strlen(args)+1);
   strcpy(renamefile,args);
   writes("305 File exists, ready for destination name\r\n");
}

void rnto(char *args)
{
   if(!checklog()) return;

   if(renamefile == NULL)
   {
      writes("503 Bad sequence of commands.\r\n");
      return;
   }
   if(rename(renamefile,args))
   {
      writes("550 rename: %s\r\n",strerror(errno));
   }
   else writes("250 RNTO command successful.\r\n");
   free(renamefile); renamefile = NULL;
}

void syst()
{
   writes("215 UNIX Type: L8\r\n");
}

void size(char *args)
{
   struct stat buf;

   if(!checklog()) return;

   if(var.trans_type == 'A')
   {
      writes("504 SIZE not implemented for type A.");
      return;
   }
   if(stat(args,&buf))
   {
      writes("550 %s: %s\r\n",args,strerror(errno));
      return;
   }
   else
   {
      writes("213 %lu\r\n",buf.st_size);
      return;
   }
}

void mdtm(char *args)
{
   struct tm *aika;
   struct stat buf; char formaatti[17];

   if(!checklog()) return;

   if(stat(args,&buf))
   {
      writes("550 %s: %s.\r\n",args,strerror(errno));
      return;
   }
   if((buf.st_mode&S_IFMT) != S_IFREG)
   {
      writes("550 %s: not a plain file.\r\n",args);
      return;
   }
   aika = gmtime(&buf.st_mtime);
   strftime(formaatti,sizeof(formaatti)-1,"%Y%m%d%H%M%S",aika);
   writes("213 %s\r\n",formaatti);
}

void dele(char *args)
{
   if(!checklog()) return;
   if(var.userclass != REAL && !checkaccess('D')) return;

   if(!remove(args))
      writes("250 DELE command successful.\r\n");
   else writes("550 %s: %s.\r\n",args,strerror(errno));
}

void cwd(char *args)
{
   int glob_opts = GLOB_NOCHECK; glob_t glt;

   if(!checklog()) return;

#ifdef GLOB_ONLYDIR
   glob_opts |= GLOB_ONLYDIR;
#endif
   glob(args,glob_opts,NULL,&glt);
   if(!glt.gl_pathc) writes("550 %s: %s.\r\n",args,strerror(ENOENT));
   else if(!chdir(glt.gl_pathv[0]))
   {
      if(!access(".message",R_OK)) viewfile(".message",250);
      writes("250 CWD command successful.\r\n");
   }
   else writes("550 %s: %s.\r\n",args,strerror(errno));
   globfree(&glt);
}

void mkd(char *args)
{
   if(!checklog()) return;
   if(var.userclass != REAL && !checkaccess('W')) return;

   if(!mkdir(args,var.cmode|0111))
      writes("250 MKD command successful.\r\n");
   else writes("550 %s: %s.\r\n",args,strerror(errno));
}

void rmd(char *args)
{
   if(!checklog()) return;
   if(var.userclass != REAL && !checkaccess('D')) return;

   if(!rmdir(args))
      writes("250 RMD command successful.\r\n");
   else writes("550 %s: %s.\r\n",args,strerror(errno));
}

void noop()
{
   writes("250 NOOP command successful.\r\n");
}

void type(char *args)
{
   switch(toupper(args[0]))
   {
      case 'A': var.trans_type = 'A'; writes("200 Type set to A.\r\n"); break;
      case 'I': var.trans_type = 'I'; writes("200 Type set to I.\r\n"); break;
      case 'E': writes("504 Type E not implemented.\r\n"); break;
      default: writes("500 'TYPE %s': Command not understood.\r\n",args);
   }
}

void port(char *args)
{
   int i,c = 0,testport; char *osoite = NULL; struct in_addr testaddr;

   for(i = 0; c < 4 && i < strlen(args); i++)
      if(args[i] == ',') c++;
   osoite = malloc(i+1); c = 0; memset(osoite,'\0',i+1);
   for(i = 0; c < 4; i++)
   {
      if(args[i] == ',')
      {
         if((++c) < 4) osoite[i] = '.';
      }
      else osoite[i] = args[i];
   }
   osoite[i] = '\0';

   if(c < 4 || !inet_aton(osoite,&testaddr))
   {
      writes("530 Illegal PORT command.\r\n");
      free(osoite); return;
   }
   c = 0;
   for(; isdigit(args[i]) && c < 5; i++) osoite[c++] = args[i]; osoite[c] = '\0';
   if(c == 5)
   {
      writes("530 Illegal PORT command.\r\n");
      free(osoite); return;
   }
   testport = atoi(osoite)*256; i++;
   for(c = 0; isdigit(args[i]) && c < 5; i++) osoite[c++] = args[i]; osoite[c] = 0;
   if(c == 5)
   {
      writes("530 Illegal PORT command.\r\n");
      free(osoite); return;
   }
   testport += atoi(osoite); data_addr.sin_addr = testaddr;
   data_addr.sin_port = htons(testport);
   writes("200 PORT command successful.\r\n"); free(osoite);
}

void setoffset(char *args)
{
   if(!isdigit(args[0]))
   {
      writes("500 'REST %s': command not understood.\r\n",args);
      return;
   }
   offset = atol(args);
   writes("350 Restarting at %s. Send STORE or RETRIEVE to initiate transfer.\r\n",args);
}

void pasv()
{
   char *a, *b; int namelen;

   if(!checklog()) return;

   if(pasv_servsockfd >= 0)
   {
      close(pasv_servsockfd); pasv_servsockfd = -1;
   }

   if( (pasv_servsockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
   {
      writes("425 Cannot open passive connection.\r\n"); return;
   }

   memset(&pasv_addr, 0, sizeof(pasv_addr));
   namelen = sizeof(pasv_addr);
   getsockname(sockfd,(struct sockaddr *)&pasv_addr,&namelen);
   pasv_addr.sin_port = 0;
   if(bind(pasv_servsockfd, (struct sockaddr *) &pasv_addr, sizeof(pasv_addr)) < 0)
   {
      writes("425 Cannot open passive connection.\r\n"); return;
   }
   listen(pasv_servsockfd,BACKLOG);
   getsockname(pasv_servsockfd,(struct sockaddr *)&pasv_addr,&namelen);
   a = (char *) &pasv_addr.sin_addr; b = (char *) &pasv_addr.sin_port;
   writes("227 Entering Passive Mode (%u,%u,%u,%u,%u,%u)\r\n"
    ,(unsigned)a[0]&0xff,(unsigned)a[1]&0xff,(unsigned)a[2]&0xff,(unsigned)a[3]&0xff
    ,(unsigned)b[0]&0xff,(unsigned)b[1]&0xff);
}

void sitehelp(char *args)
{
   int i,j,rowlen = 8;

   if(args == NULL)
   {
      writes("214-Recognised SITE commands:\r\n");
      writes("214-HELP    ");
      for(i=0;i<cfg.sitecmdcount;i++)
         if(cfg.sitecmds[i].avail&var.userclass)
         {
            if(rowlen >= 52) { writes("\r\n214-"); rowlen = 0; }
            writes(cfg.sitecmds[i].name); rowlen += strlen(cfg.sitecmds[i].name);
            for(j=4-strlen(cfg.sitecmds[i].name)%4;j;j--) { writes(" "); rowlen++; }
            if(strlen(cfg.sitecmds[i].name) < 4) { writes("    "); rowlen += 4; }
         }
      writes("\r\n214 End of SITE command list\r\n");
      return;
   }
   for(i=0;i<cfg.sitecmdcount;i++)
           if(!strcasecmp(args,cfg.sitecmds[i].name))
            {
               writes(cfg.sitecmds[i].helptxt); return;
            }
   if(!strcasecmp(args,"HELP"))
   {
      writes("214 SITE HELP [command] : retrieve information of a specific command\r\n");
      return;
   }
   writes("502 Unknown command "); writes(args); writes(".\r\n");
}

void site(char *line)
{
   int i; char *cmd;

   if(!checklog()) return;

   cmd = strsep(&line," ");
   if(!cmd)
   {
      writes("550 'SITE': command not understood."); return;
   }
   for(i = 0; i < strlen(cmd); i++) cmd[i] = toupper(cmd[i]);

   if(cfg.sitecmds) for(i=0;i<cfg.sitecmdcount;i++)
   {
      if(!strcmp(cmd,cfg.sitecmds[i].name))
      {
         cfg.sitecmds[i].cmd(line); return;
      }
   }
   if(!strcmp(cmd,"HELP"))
   {
      sitehelp(line); return;
   }
   writes("500 'SITE %s",cmd);
   if(line != NULL) writes(" %s",line);
   writes("': command not understood.\r\n");
}

void helpcmd()
{
   writes("214-Recognized commands:\r\n");
   writes("214-USER    CWD    PORT    MKD    MODE    RETR    MODE    ABOR\r\n");
   writes("214-PASS    CDUP   RETR    RMD    NLST    TYPE    HELP    PASV\r\n");
   writes("214 QUIT    LIST   STOR    PWD    CDUP    SIZE    MDTM\r\n");
}

void cmdloop()
{
   int i; char line[MAXLINE], *cmd = NULL, *args = NULL;

   for(;;)
   {
      memset(line,0,MAXLINE); readline(sockfd,line);
      update_env();
      if(strncasecmp(line,"PASS",4))
      {
         ftplog("CMD",line); activity(line);
      }
      else
      {
         ftplog("CMD","PASS xxxxxxx"); activity("PASS xxxxxxx");
      }
      tildeexp(line);
      args = line; cmd = strsep(&args," ");
      for(i = 0; i < strlen(cmd); i++) cmd[i] = toupper(cmd[i]);

      if(!var.logged && !strcmp(cmd,"AUTH")) auth(args);
      else if(!strcmp(cmd,"SYST")) syst();
      else if(args != NULL && !strcmp(cmd,"USER")) user(args);
      else if(args != NULL && !strcmp(cmd,"PASS")) pass(args);
      else if(!strcmp(cmd,"NOOP")) noop();
      else if(args != NULL && !strcmp(cmd,"PORT")) port(args);
      else if(!strcmp(cmd,"QUIT")) quit();
      else if(args != NULL && !strcmp(cmd,"TYPE")) type(args);
      else if(!strcmp(cmd,"PWD")) pwd();
      else if(args != NULL && !strcmp(cmd,"CWD")) cwd(args);
      else if(!strcmp(cmd,"CDUP")) cwd("..");
      else if(args != NULL && !strcmp(cmd,"MKD")) mkd(args);
      else if(args != NULL && !strcmp(cmd,"RMD")) rmd(args);
      else if(args != NULL && !strcmp(cmd,"DELE")) dele(args);
      else if(!strcmp(cmd,"LIST")) listcmd(0,args);
      else if(!strcmp(cmd,"NLST")) listcmd(1,args);
      else if(args != NULL && !strcmp(cmd,"RETR")) get(args);
      else if(args != NULL && !strcmp(cmd,"STOR")) put(args);
      else if(args != NULL && !strcmp(cmd,"REST")) setoffset(args);
      else if(args != NULL && !strcmp(cmd,"SITE")) site(args);
      else if(args != NULL && !strcmp(cmd,"SIZE")) size(args);
      else if(!strcmp(cmd,"HELP")) helpcmd();
      else if(args != NULL && !strcmp(cmd,"MDTM")) mdtm(args);
      else if(args != NULL && !strcmp(cmd,"RNFR")) rnfr(args);
      else if(args != NULL && !strcmp(cmd,"RNTO")) rnto(args);
      else if(!strcmp(cmd,"PASV")) pasv();
      else writes("502 '%s': command not understood.\r\n",cmd);
      activity(NULL);
   }
}
