#include <stdio.h>
#define __USE_XOPEN
#include <unistd.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/file.h>
#include <string.h>
#include <shadow.h>
#include <pwd.h>
#include <grp.h>
#include "dftpd.h"

/* libdftp dependencies */

int dftp_ugetbyname(int, char *, struct dftp_uent *);
void dftp_ufree(struct dftp_uent *);
int dftp_ggetbyname(int, char *, struct dftp_gent *);
int dftp_ggetbyid(int, unsigned int, struct dftp_gent *);
void dftp_gfree(struct dftp_gent *);
int dftp_utmpgetent(int, struct dftp_utmpent *);

/* aux.c dependencies */

short gethname(struct sockaddr_in *, char *);
void viewfile(const char *, const short);

/* srvcommon.c dependencies */

void writes(const char *, ...);
int checkidentip(struct sockaddr_in *, int);

/* dftpd.c dependencies */

void ftplog(const char *, const char *, ...);

extern struct dftp_config cfg;
extern struct dftp_vars var;
extern int passwdfile, groupfile, utmpfile;
extern struct dftp_utmpent cur_utmp;
extern struct sockaddr_in cli_addr;

/* This function is for bnc4all support */

void auth(char *args)
{
   char *ptr, *ptr2;

   if((ptr=strchr(args,' '))) *ptr = 0;
   ptr = args; free(var.ident);
   if((ptr2=strsep(&ptr,"@")) && strcmp(ptr2,"(null)"))
   {
      var.ident = malloc(strlen(ptr2)+1);
      strcpy(var.ident,ptr2);
   }
   else var.ident = NULL;
   inet_aton(ptr,&cli_addr.sin_addr);
   gethname(&cli_addr,cur_utmp.controlhostname);
}

void loginfail()
{
   var.logged = 0; var.loginfails++;
   if(var.loginfails == cfg.maxloginfails)
   {
      writes("531 Excess login failures. Closing control connection.\r\n");
      exit(0);
   }
}

void user(char *args)
{
   if(var.logged == 2)
   {
      writes("530 Already logged in.\r\n"); return;
   }

   if(!strcmp(args,"anonymous") || !strcmp(args,"ftp"))
   {
      strcpy(var.user.name,"ftp");
      writes("331 Guest login ok, send your complete e-mail address as password.\r\n");
   }
   else
   {
      strncpy(var.user.name,args,NAMELEN);
      writes("331 Password required for %s.\r\n",var.user.name);
   }
   var.logged = 1;
}

int auth_real(char *password)
{
   struct passwd *user; struct group *group;
   struct spwd *sh_user;

   if(!(user = getpwnam(var.user.name))) return -2;
   if(cfg.shadow)
   {
      lckpwdf();
      sh_user = getspnam(var.user.name);
      ulckpwdf();
      if(!sh_user) return -2;
      if(strcmp(crypt(password,sh_user->sp_pwdp),sh_user->sp_pwdp)) return -1;
   }
   else if(strcmp(crypt(password,user->pw_passwd),user->pw_passwd)) return -1;
   if(!(group = getgrgid(user->pw_gid))) return 0;
   var.user.uid = user->pw_uid;
   var.user.gid = user->pw_gid;
   var.user.homedir = malloc(strlen(user->pw_dir)+1);
   strcpy(var.user.homedir,user->pw_dir);
   var.userclass = REAL; return 1;
}

int auth_guest(char *password)
{
   if(!dftp_ugetbyname(passwdfile,var.user.name,&var.user)) return -2;
   if(strcmp(crypt(password,var.user.passwd),var.user.passwd))
   {
      dftp_ufree(&var.user); return -1;
   }
   if(!dftp_ggetbyid(groupfile,var.user.gid,&var.group))
   {
      dftp_ufree(&var.user); return 0;
   }
   var.userclass = GUEST; return 1;
}

int auth_anon(char *passwd)
{
   if(!dftp_ugetbyname(passwdfile,"ftp",&var.user)) return -2;
   if(!dftp_ggetbyname(groupfile,"ftp",&var.group))
   {
      dftp_ufree(&var.user); return 0;
   }
   var.userclass = ANONYMOUS; return 1;
}

void pass(char *args)
{
   int i,success; struct dftp_utmpent entry;

   if(var.logged != 1)
   {
      writes("530 Login with USER first.\r\n"); return;
   }

   if(!strcmp(var.user.name,"ftp")) 
   {
      if(auth_anon(args) < 1)
      {
         writes("530 Anonymous access is disabled.\r\n"); exit(0);
      }
      success = 1;
   }
   else success = auth_guest(args);
   switch(success)
   {
      case -2: if(cfg.realusersupport) break;
               ftplog("AUTH","Login failed - user not found"); loginfail();
               writes("530 Login incorrect.\r\n"); return;
      case -1: ftplog("AUTH","Login failed - incorrect password"); loginfail();
               writes("530 Login incorrect.\r\n"); return;
      case  0: ftplog("AUTH","Login failed - primary group not found"); loginfail();
               writes("530 Login incorrect.\r\n"); return;
   }
   if(success == -2) switch(auth_real(args))
   {
      case -2: ftplog("AUTH","Login failed - user not found"); loginfail();
               writes("530 Login incorrect.\r\n"); return;
      case -1: ftplog("AUTH","Login failed - incorrect password"); loginfail();
               writes("530 Login incorrect.\r\n"); return;
      case  0: ftplog("AUTH","Login failed - primary group not found"); loginfail();
               writes("530 Login incorrect.\r\n"); return;
   }
   if(var.userclass != REAL)
   {
      if(!var.user.maxconn)
      {
         ftplog("AUTH","Login failed - account is disabled");
         if(var.userclass != ANONYMOUS)
            writes("530 Your account is disabled. Closing control connection.\r\n");
         else writes("530 Anonymous access is disabled.\r\n");
         exit(0);
      }
      switch(checkidentip(&cli_addr,ID_CONTROL))
      {
         case -1: ftplog("AUTH","Login failed - no matching ident@ip combinations found");
                 writes("530 Login incorrect.\r\n"); loginfail(); return;
         case 1: ftplog("AUTH","Login failed - ident@ip combination is banned");
                 writes("530 Login incorrect.\r\n"); loginfail(); return;
      }
      flock(utmpfile,LOCK_EX); i = 0; lseek(utmpfile,0,SEEK_SET);
      while( (success=dftp_utmpgetent(utmpfile,&entry)) )
         if(success < 0 && !strcmp(entry.name,var.user.name)) i++;
      flock(utmpfile,LOCK_UN);

      if(i >= var.user.maxconn)
      {
         if(cfg.debuglevel >= 1) ftplog("AUTH","DEBUG: login failed - maximum connection amount reached");
         writes("430 Maximum connection amount of %i reached. Closing control connection.\r\n",var.user.maxconn);
         exit(0);
      }
   }

   /* User & password ok, initialize user */

   if(var.userclass == REAL)
   {
      initgroups(var.user.name,var.user.gid);
      setpwent(); setgrent();
   }
   if(var.user.homedir && strstr(var.user.homedir,"/./"))
   {
      char *tempdir;
   
      *strstr(var.user.homedir,"/./") = 0;
      chroot(var.user.homedir);
      tempdir = malloc(strlen(var.user.homedir+strlen(var.user.homedir)+2)+1);
      strcpy(tempdir,var.user.homedir+strlen(var.user.homedir)+2);
      free(var.user.homedir); var.user.homedir = tempdir;
   }
   else if(var.userclass != REAL)
   {
      if(chroot(cfg.sitedir))
      {
         ftplog("ERR","Cannot CHROOT to %s",cfg.sitedir);
         writes("552-Cannot CHROOT to base directory.\r\n");
         writes("552 Closing connection.\r\n");
         exit(3);
      }
      initgroups(var.user.name,var.user.gid);
   }
   if(var.user.homedir) chdir(var.user.homedir);
   var.logged = 2; if(!access(cfg.welcomefile,R_OK)) viewfile(cfg.welcomefile,230);
   switch(var.userclass)
   {
      case ANONYMOUS: if(args[strlen(args)-1] == '@')
                      ftplog("AUTH","Anonymous login from %s%s",args,cur_utmp.controlhostname);
                      else ftplog("AUTH","Anonymous login from %s",args);
                      writes("230 Guest login ok.\r\n"); break;
      case     GUEST: ftplog("AUTH","Guest login - %s",var.user.name);
                      writes("230 User %s logged in.\r\n",var.user.name); break;
      case      REAL: ftplog("AUTH","User login - %s",var.user.name);
                      writes("230 User %s logged in.\r\n",var.user.name); break;
   }
   setegid(var.group.id); seteuid(var.user.uid);
   strcpy(cur_utmp.name,var.user.name);
}
